@@ -721,6 +721,11 @@ void WaterTable2::setElevationRange(WaterTable2::Scalar newMin,WaterTable2::Scal
 	domain.max[2]=newMax;
 	}
 
+void WaterTable2::setAttenuation(GLfloat newAttenuation)
+	{
+	attenuation=newAttenuation;
+	}
+
 void WaterTable2::setMaxStepSize(GLfloat newMaxStepSize)
 	{
 	maxStepSize=newMaxStepSize;
